import { assertDefined } from "@rickosborne/guard";
export const withNullsFirst = (comparator) => {
    return function nullsFirst(a, b) {
        if (a == null && b == null) {
            return 0;
        }
        else if (a == null) {
            return -1;
        }
        else if (b == null) {
            return 1;
        }
        return comparator(a, b);
    };
};
export const withNullsLast = (comparator) => {
    return function nullsLast(a, b) {
        if (a == null && b == null) {
            return 0;
        }
        else if (a == null) {
            return 1;
        }
        else if (b == null) {
            return -1;
        }
        return comparator(a, b);
    };
};
const withDesc = (comparator) => function desc(a, b) {
    return comparator(b, a);
};
export const chainableComparator = (comparator) => {
    let _nullsFirst;
    let _nullsLast;
    let _desc;
    const added = ["nullsFirst", "nullsLast", "desc"];
    return new Proxy(comparator, {
        get(target, p) {
            if (p === "nullsFirst") {
                _nullsFirst ??= chainableComparator(withNullsFirst(comparator));
                return _nullsFirst;
            }
            if (p === "nullsLast") {
                _nullsLast ??= chainableComparator(withNullsLast(comparator));
                return _nullsLast;
            }
            if (p === "desc") {
                _desc ??= chainableComparator(withDesc(comparator));
                return _desc;
            }
            return target[p];
        },
        getOwnPropertyDescriptor(target, p) {
            if (added.includes(p)) {
                return {
                    enumerable: true,
                    get: () => target[p],
                };
            }
            return Object.getOwnPropertyDescriptor(target, p);
        },
        has(target, p) {
            return added.includes(p) || Object.hasOwn(target, p);
        },
        ownKeys(target) {
            return [...added, ...Object.keys(target)];
        },
    });
};
export const numberAsc = (a, b) => a - b;
export const comparatorBuilder = () => {
    const ops = [];
    const withPrevious = (modifier) => {
        const lastIndex = ops.length - 1;
        const comparator = ops[lastIndex][1];
        assertDefined(comparator, "Prior comparator");
        ops[lastIndex][1] = modifier(comparator);
    };
    const builder = {
        build() {
            return (a, b) => {
                for (const [accessor, comparator] of ops) {
                    const aValue = accessor(a);
                    const bValue = accessor(b);
                    const result = comparator(aValue, bValue);
                    if (result !== 0) {
                        return result;
                    }
                }
                return 0;
            };
        },
        get desc() {
            withPrevious((c) => withDesc(c));
            return builder;
        },
        get nullsFirst() {
            withPrevious((c) => withNullsFirst(c));
            return builder;
        },
        get nullsLast() {
            withPrevious((c) => withNullsLast(c));
            return builder;
        },
        num(accessor) {
            ops.push([accessor, (a, b) => {
                    if (a == null && b == null) {
                        return 0;
                    }
                    if (a == null) {
                        return 1;
                    }
                    if (b == null) {
                        return -1;
                    }
                    return a - b;
                }]);
            return builder;
        },
        str(accessor) {
            ops.push([accessor, (a, b) => {
                    if (a == null && b == null) {
                        return 0;
                    }
                    if (a == null) {
                        return 1;
                    }
                    if (b == null) {
                        return -1;
                    }
                    return a.localeCompare(b);
                }]);
            return builder;
        },
    };
    return builder;
};
//# sourceMappingURL=comparator.js.map